@@ -7,7 +7,34 @@ from types import MethodType
 import torch
 
 
-def simple2log(x):
+def covariance(x, y):
+    """Compute covariance between two 2D tensors.
+
+    Parameters
+    ----------
+    x : torch.tensor
+        Torch tensor of shape `(n_samples, horizon)`
+
+    y : torch.tensor
+        Tensor of shape `(n_samples, horizon)`
+
+    Returns
+    -------
+    cov : torch.tensor
+        Torch tensor of shape `(n_samples,)`.
+    """
+    n_samples, horizon = x.shape
+    mean_x = x.mean(dim=1, keepdim=True)
+    mean_y = y.mean(dim=1, keepdim=True)
+    xm = x - mean_x  # (n_samples, horizon)
+    ym = y - mean_y  # (n_samples, horizon)
+
+    cov = (xm * ym).sum(dim=1) / horizon
+
+    return cov
+
+
+def log2simple(x):
     """Turn simple returns into log returns.
 
     r_simple = exp(r_log) - 1.
@@ -26,7 +53,7 @@ def simple2log(x):
     return torch.exp(x) - 1
 
 
-def log2simple(x):
+def simple2log(x):
     """Turn log returns into simple returns.
 
     r_log = ln(r_simple + 1).
@@ -45,7 +72,7 @@ def log2simple(x):
     return torch.log(x + 1)
 
 
-def portfolio_returns(weights, y, input_type='log', output_type='simple'):
+def portfolio_returns(weights, y, input_type='log', output_type='simple', rebalance=False):
     """Compute portfolio returns.
 
     Parameters
@@ -62,10 +89,14 @@ def portfolio_returns(weights, y, input_type='log', output_type='simple'):
     output_type : str, {'log', 'simple'}
         What type of returns are we dealing with in the output.
 
+    rebalance : bool
+        If True, each timestep the weights are adjusted to be equal to be equal to the original ones. Note that
+        this assumes that we tinker with the portfolio. If False, the portfolio evolves untouched.
+
     Returns
     -------
     portfolio_returns : torch.Tensor
-        Of shape (n_samples, horizon)
+        Of shape (n_samples, horizon) representing per timestep portfolio returns.
 
     """
     if input_type == 'log':
@@ -79,12 +110,13 @@ def portfolio_returns(weights, y, input_type='log', output_type='simple'):
 
     n_samples, horizon, n_assets = simple_returns.shape
 
-    res = []
+    weights_ = weights.view(n_samples, 1, n_assets).repeat(1, horizon, 1)  # (n_samples, horizon, n_assets)
 
-    for i in range(n_samples):
-        res.append(simple_returns[i] @ weights[i])  # (horizon, n_assets)x(n_assets)=(horizon,)
+    if not rebalance:
+        weights_unscaled = (1 + simple_returns).cumprod(1)[:, :-1, :] * weights_[:, 1:, :]
+        weights_[:, 1:, :] = weights_unscaled / weights_unscaled.sum(2, keepdim=True)
 
-    out = torch.stack(res, dim=0)
+    out = (simple_returns * weights_).sum(-1)
 
     if output_type == 'log':
         return simple2log(out)
@@ -96,7 +128,7 @@ def portfolio_returns(weights, y, input_type='log', output_type='simple'):
         raise ValueError('Unsupported output type: {}'.format(output_type))
 
 
-def portfolio_cumulative_returns(weights, y, input_type='log', output_type='simple'):
+def portfolio_cumulative_returns(weights, y, input_type='log', output_type='simple', rebalance=False):
     """Compute cumulative portfolio returns.
 
     Parameters
@@ -114,14 +146,18 @@ def portfolio_cumulative_returns(weights, y, input_type='log', output_type='simp
     output_type : str, {'log', 'simple'}
         What type of returns are we dealing with in the output.
 
+    rebalance : bool
+        If True, each timestep the weights are adjusted to be equal to be equal to the original ones. Note that
+        this assumes that we tinker with the portfolio. If False, the portfolio evolves untouched.
+
     Returns
     -------
     torch.Tensor
         Tensor of shape `(n_samples, horizon)`.
 
     """
-    prets = portfolio_returns(weights, y, input_type=input_type, output_type='log')  # we can aggregate overtime by sum
-    log_prets = torch.cumsum(prets, dim=1)
+    prets = portfolio_returns(weights, y, input_type=input_type, output_type='log', rebalance=rebalance)
+    log_prets = torch.cumsum(prets, dim=1)  # we can aggregate log returns over time by sum
 
     if output_type == 'log':
         return log_prets
@@ -328,6 +364,125 @@ class Loss:
             raise TypeError('Unsupported type: {}'.format(type(power)))
 
 
+class Alpha(Loss):
+    """Negative alpha with respect to a selected portfolio.
+
+    Parameters
+    ----------
+    benchmark_weights : torch.tensor or None
+        Weights of the benchmark portfolio of shape `(n_assets,). Note that this loss assumes it will be always located
+        under this index in the `y` tensor. If None then equally weighted portfolio.
+
+    returns_channel : int
+        Which channel of the `y` target represents returns.
+
+    input_type : str, {'log', 'simple'}
+        What type of returns are we dealing with in `y`.
+
+    """
+
+    def __init__(self, benchmark_weights=None, returns_channel=0, input_type='log'):
+        self.benchmark_weights = benchmark_weights
+        self.returns_channel = returns_channel
+        self.input_type = input_type
+
+    def __call__(self, weights, y):
+        """Compute negative alpha with respect to the benchmark portfolio.
+
+        Parameters
+        ----------
+        weights : torch.Tensor
+            Tensor of shape `(n_samples, n_assets)` representing the predicted weights by our portfolio optimizer.
+
+        y : torch.Tensor
+            Tensor of shape `(n_samples, n_channels, horizon, n_assets)` representing the evolution over the next
+            `horizon` timesteps.
+
+        Returns
+        -------
+        torch.Tensor
+            Tensor of shape `(n_samples,)` representing the per sample negative alpha.
+
+        """
+        n_samples, n_assets = weights.shape
+        device, dtype = weights.device, weights.dtype
+        portfolio_rets = portfolio_returns(weights,
+                                           y[:, self.returns_channel, ...],
+                                           input_type=self.input_type,
+                                           output_type='simple')  # (n_samples, horizon)
+
+        if self.benchmark_weights is None:
+            benchmark_weights = torch.ones(n_samples, n_assets, dtype=dtype, device=device) / n_assets
+        else:
+            benchmark_weights = self.benchmark_weights[None, :].repeat(n_samples, 1).to(device=device, dtype=dtype)
+
+        benchmark_rets = portfolio_returns(benchmark_weights,
+                                           y[:, self.returns_channel, ...],
+                                           input_type=self.input_type,
+                                           output_type='simple')  # (n_samples, horizon)
+
+        cov = covariance(benchmark_rets, portfolio_rets)
+        beta = cov / benchmark_rets.var(dim=1)
+        alpha = portfolio_rets.mean(dim=1) - beta * benchmark_rets.mean(dim=1)
+
+        return -alpha
+
+    def __repr__(self):
+        """Generate representation string."""
+        return "{}(benchmark_weights={},returns_channel={}, input_type='{}')".format(self.__class__.__name__,
+                                                                                     self.benchmark_weights,
+                                                                                     self.returns_channel,
+                                                                                     self.input_type)
+
+
+class CumulativeReturn(Loss):
+    """Negative cumulative returns.
+
+    Parameters
+    ----------
+    returns_channel : int
+        Which channel of the `y` target represents returns.
+
+    input_type : str, {'log', 'simple'}
+        What type of returns are we dealing with in `y`.
+    """
+
+    def __init__(self, returns_channel=0, input_type='log'):
+        self.returns_channel = returns_channel
+        self.input_type = input_type
+
+    def __call__(self, weights, y):
+        """Compute negative simple cumulative returns.
+
+        Parameters
+        ----------
+        weights : torch.Tensor
+            Tensor of shape `(n_samples, n_assets)` representing the predicted weights by our portfolio optimizer.
+
+        y : torch.Tensor
+            Tensor of shape `(n_samples, n_channels, horizon, n_assets)` representing the evolution over the next
+            `horizon` timesteps.
+
+        Returns
+        -------
+        torch.Tensor
+            Tensor of shape `(n_samples,)` representing the per sample negative simple cumulative returns.
+
+        """
+        crets = portfolio_cumulative_returns(weights,
+                                             y[:, self.returns_channel, ...],
+                                             input_type=self.input_type,
+                                             output_type='simple')
+
+        return -crets[:, -1]
+
+    def __repr__(self):
+        """Generate representation string."""
+        return "{}(returns_channel={}, input_type='{}')".format(self.__class__.__name__,
+                                                                self.returns_channel,
+                                                                self.input_type)
+
+
 class LargestWeight(Loss):
     """Largest weight loss.
 
@@ -362,6 +517,52 @@ class LargestWeight(Loss):
         return "{}()".format(self.__class__.__name__)
 
 
+class MaximumDrawdown(Loss):
+    """Negative of the maximum drawdown."""
+
+    def __init__(self, returns_channel=0, input_type='log'):
+        self.returns_channel = returns_channel
+        self.input_type = input_type
+
+    def __call__(self, weights, y):
+        """Compute maximum drawdown.
+
+        Parameters
+        ----------
+        weights : torch.Tensor
+            Tensor of shape `(n_samples, n_assets)` representing the predicted weights by our portfolio optimizer.
+
+        y : torch.Tensor
+            Tensor of shape `(n_samples, n_channels, horizon, n_assets)` representing the evolution over the next
+            `horizon` timesteps.
+
+        Returns
+        -------
+        torch.Tensor
+            Tensor of shape `(n_samples,)` representing the per sample maximum drawdown.
+
+        """
+        cumrets = 1 + portfolio_cumulative_returns(weights,
+                                                   y[:, self.returns_channel, ...],
+                                                   input_type=self.input_type,
+                                                   output_type='simple')
+
+        cummax = torch.cummax(cumrets, 1)[0]  # (n_samples, n_timesteps)
+
+        div = (cumrets / cummax) - 1  # (n_samples, n_timesteps)
+
+        end = div.argmin(dim=1)  # (n_samples,)
+        mdd = div.gather(1, end.view(-1, 1)).view(-1)
+
+        return -mdd
+
+    def __repr__(self):
+        """Generate representation string."""
+        return "{}(returns_channel={}, input_type='{}')".format(self.__class__.__name__,
+                                                                self.returns_channel,
+                                                                self.input_type)
+
+
 class MeanReturns(Loss):
     """Negative mean returns."""
 
@@ -403,13 +604,78 @@ class MeanReturns(Loss):
                                                                                   self.output_type)
 
 
-class SharpeRatio(Loss):
-    """Negative Sharpe ratio."""
+class Quantile(Loss):
+    """Compute negative percentile.
 
-    def __init__(self, returns_channel=0, input_type='log', output_type='simple'):
+    Parameters
+    ----------
+    q : float
+        Number from (0, 1) representing the quantile.
+
+    """
+
+    def __init__(self, returns_channel=0, q=0.1):
+        self.returns_channel = returns_channel
+        self.q = q
+
+    def __call__(self, weights, y):
+        """Compute negative quantile.
+
+        Parameters
+        ----------
+        weights : torch.Tensor
+            Tensor of shape `(n_samples, n_assets)` representing the predicted weights by our portfolio optimizer.
+
+        y : torch.Tensor
+            Tensor of shape `(n_samples, n_channels, horizon, n_assets)` representing the evolution over the next
+            `horizon` timesteps.
+
+        Returns
+        -------
+        torch.Tensor
+            Tensor of shape `(n_samples,)` representing the per sample negative quantile.
+
+        """
+        prets = portfolio_returns(weights,
+                                  y[:, self.returns_channel, ...])  # (n_samples, horizon)
+
+        _, horizon = prets.shape
+
+        k = 1 + round(self.q * (horizon - 1))
+        return -prets.kthvalue(k)[0]
+
+    def __repr__(self):
+        """Generate representation string."""
+        return "{}(returns_channel={})".format(self.__class__.__name__, self.returns_channel)
+
+
+class SharpeRatio(Loss):
+    """Negative Sharpe ratio.
+
+    Parameters
+    ----------
+    rf : float
+        Risk-free rate.
+
+    returns_channel : int
+        Which channel of the `y` target represents returns.
+
+    input_type : str, {'log', 'simple'}
+        What type of returns are we dealing with in `y`.
+
+    output_type : str, {'log', 'simple'}
+        What type of returns are we dealing with in the output.
+
+    eps : float
+        Additional constant added to the denominator to avoid division by zero.
+    """
+
+    def __init__(self, rf=0, returns_channel=0, input_type='log', output_type='simple', eps=1e-4):
+        self.rf = rf
         self.returns_channel = returns_channel
         self.input_type = input_type
         self.output_type = output_type
+        self.eps = eps
 
     def __call__(self, weights, y):
         """Compute negative sharpe ratio.
@@ -434,14 +700,17 @@ class SharpeRatio(Loss):
                                   input_type=self.input_type,
                                   output_type=self.output_type)
 
-        return -prets.mean(dim=1) / prets.std(dim=1)
+        return -(prets.mean(dim=1) - self.rf) / (prets.std(dim=1) + self.eps)
 
     def __repr__(self):
         """Generate representation string."""
-        return "{}(returns_channel={}, input_type='{}', output_type='{}')".format(self.__class__.__name__,
-                                                                                  self.returns_channel,
-                                                                                  self.input_type,
-                                                                                  self.output_type)
+        return "{}(rf={}, returns_channel={}, input_type='{}', output_type='{}', eps={})".format(
+            self.__class__.__name__,
+            self.rf,
+            self.returns_channel,
+            self.input_type,
+            self.output_type,
+            self.eps)
 
 
 class Softmax(Loss):
@@ -478,12 +747,32 @@ class Softmax(Loss):
 
 
 class SortinoRatio(Loss):
-    """Negative Sortino ratio."""
+    """Negative Sortino ratio.
 
-    def __init__(self, returns_channel=0, input_type='log', output_type='simple'):
+    Parameters
+    ----------
+    rf : float
+        Risk-free rate.
+
+    returns_channel : int
+        Which channel of the `y` target represents returns.
+
+    input_type : str, {'log', 'simple'}
+        What type of returns are we dealing with in `y`.
+
+    output_type : str, {'log', 'simple'}
+        What type of returns are we dealing with in the output.
+
+    eps : float
+        Additional constant added to the denominator to avoid division by zero.
+    """
+
+    def __init__(self, rf=0, returns_channel=0, input_type='log', output_type='simple', eps=1e-4):
+        self.rf = rf
         self.returns_channel = returns_channel
         self.input_type = input_type
         self.output_type = output_type
+        self.eps = eps
 
     def __call__(self, weights, y):
         """Compute negative Sortino ratio of portfolio return over the horizon.
@@ -508,14 +797,17 @@ class SortinoRatio(Loss):
                                   input_type=self.input_type,
                                   output_type=self.output_type)
 
-        return -prets.mean(dim=1) / (torch.sqrt(torch.mean(torch.relu(-prets) ** 2, dim=1)) + 1e-6)
+        return -(prets.mean(dim=1) - self.rf) / (torch.sqrt(torch.mean(torch.relu(-prets) ** 2, dim=1)) + self.eps)
 
     def __repr__(self):
         """Generate representation string."""
-        return "{}(returns_channel={}, input_type='{}', output_type='{}')".format(self.__class__.__name__,
-                                                                                  self.returns_channel,
-                                                                                  self.input_type,
-                                                                                  self.output_type)
+        return "{}(rf={}, returns_channel={}, input_type='{}', output_type='{}', eps={})".format(
+            self.__class__.__name__,
+            self.rf,
+            self.returns_channel,
+            self.input_type,
+            self.output_type,
+            self.eps)
 
 
 class SquaredWeights(Loss):
